#!/usr/bin/perl


# Author: Martin Fuerstenau
#         martin.fuerstenau_at_oce.com or Martin.fuerstenau_at_nagiossw.org
#
# Date:   06 May 2012
# 
# Purpose
#
# - Submitting alert emails to Nagios,Icinga,.....
#
# Features of the program:
#
# - Accepts Internet mail (text), MIME and Multipart MIME messages (nested MIME not supported)
# - Extracts sender, subject and content (text)
# - Converts HTML in message content to text
# - Converts base64 coded message content to text
# - Converts german umlauts to ASCII
# - Has a API to add filter scripts written in Perl or any other language to 
#   extract and reformat
#   - host and/or application from subject and/or content 
#   - data from subject and/or content 
#   - reformat data and deliver the data back to the program
# - Writes the reformatted data (host, errorcode, content) to the Nagios pipe
# - Includes a option to store mails in a given directory (because you need some stored mails to develop your filter)
# - Includes a testmode for testing filters while developing them.
#
#
# History and Changes:
# 
# - 06 May 2012 Version 1
#    - First released version


use Mail::Internet;
use Mail::Header;
use Mail::Audit;
use MIME::Entity;
use MIME::Words qw(:all);
use MIME::Base64;
use Getopt::Long;
use Date::Calc qw(:all);
use DBI;
use Config::IniFiles;

use Lingua::DE::ASCII;                                    # Only usefull for Germany. Translates umlauts.
 
#--- Start presets and declarations ----------------------------------------------------------------

my $progname = "mail2mon";                                # Name of program
my $noa = "";                                             # Number of arguments handled over
                                                          # the program
my $help = "";                                            # For printing the help message

my $content = "";                                         # The cleaned and prepareted content of the message
my $content_logsize = "250";                              # How much characters from the content will be
                                                          # written to the logfile in case of trouble
                                                          #.
                                                          # BEWARE!!
                                                          #
                                                          # If empty the whole content will be written to log

my $mail = "";                                            # The mail itself
my $mail2store = "";                                      # A copy of the mail itself for storage
my $tom = "";                                             # tom (type of mail) contains ....
my $head = "";                                            # Contains the head
my $sender = "";                                          # Contains the sender
my $subject = "";                                         # Contains the subject
my $bodyh = "";                                           # With MIME we need the body handle
my $type = "";                                            # The intended MIME type (as declared in the header of the mail)
my $num_parts = "";                                       # The number of parts the mail contains if it is a MIME mail.
my $parts_counter = "";                                   # Needed only as a counter in a for loop

my $content_type = "";                                    # The effective MIME type (in case decoding failed)
my $content_transfer_encoding = "";                       # 7-bit, base64...
my $content_length = "1000";                              # Maximum content lenght. If it is longer .... CUT

my $header = "";                                          # Opposite to the body the complete header of a MIME message
my $parts_header = "";                                    # Same as header but for parts
my $tmpmail = "";                                         # Only a container for work
my $charset = "";                                         # The encoded character set - needed for decoding

my $logfile = "";                                         # logfile 
my $logfile_default = "/var/log/mail2mon/mail2mon.log";   # Default logfile

my %logmail;                                              # Contains te values from the configfile
                                                          # regarding mail logging. See configfile and readme.

my $m2mon_dir= "/usr/lib/nagios/mail2mon/";
my $m2mon_cfg = $m2mon_dir."conf/mail2mon.cfg";           # Central configuration file
my $m2mon_cfg_t = $m2mon_dir."conf/mail2mon_t.cfg";       # Central configuration file for testmode.
                                                          # These 2 files are necessary for testmode. If you use the same 
                                                          # configuration file as for production it may happen that all 
                                                          # emails for the newly added host/application will enter the
                                                          # monitor system even if the filter is not fully developed.

my $config;                                               # Needed for Config::IniFiles as a handle


my $sending_host = "";                                    # Host sending the mail (necessary)
my $sending_application = "";                             # Application sending the mail (optional)

my $filter_dir = $m2mon_dir."filter";                     # The name of the directoriy where the filters resides
my $filter = "";                                          # Shout contain the name of the filter
my $filter_default = "default.fltr";                      # Set the default filter. If empty no default will be used.
my %filter_return;                                        # This is the array containg the returnvalues 
                                                          # from the filter

my $multiline;                                            # Multiline output in overview. This mean technically that
                                                          # a multiline output uses a HTML <br> for the GUI instead of
                                                          # Be aware that your messing connections (email, SMS...) must use
                                                          # a filter to file out the <br>. A sed oneliner like the following
                                                          # will do the job:
                                                          # sed 's/<[^<>]*>//g'
my $multiline_def = " ";                                   # Default for $multiline;

my $nagios_cmd = "/var/spool/nagios/nagios.cmd";          # Nagios command interface
my $epochtime = time;


# First we have to fix  the number of arguments

$noa=$#ARGV;

# Now the switches for help or testing

Getopt::Long::Configure('bundling');
GetOptions
	("h"   => \$help,                 "help"               => \$help,
	 "l"   => \$logfile,              "logfile"            => \$logfile,
	 "t"   => \$testmode,             "test"               => \$testmode);


# 1. Only help?

if ($help)
   {
   print_help();
   exit 0;
   }

# Right number of arguments (therefore NOA :-)) )

if ( $noa >= 3 )
   {
   print_usage();
   exit 1;
   }

# So check for the logfile - default or something special?

if (!$logfile)
   {
   $logfile=$logfile_default;
   }


# If we have set the system to testmode we have to use the test config file instead of the production one

if ($testmode)
   {
   $m2mon_cfg = $m2mon_cfg_t;
   }


# And now we have to check for the configuration file and also check for a correct syntax

if( open(MAIL_TO_MON,"<$m2mon_cfg") )
  {
  close(MAIL_TO_MON);

  $config = new Config::IniFiles( -file => $m2mon_cfg);
      
  if(!$config)
    {
    write_log("Error in config file - please check the syntax in the config file");
    exit;
    }
  }   
else
  {
  write_log("Error - configuration file not found");
  exit;
  }


# Mail::Audit makes it easy to seperate a MailInternet from a MimeEntity

$mail = Mail::Audit->new(log => $logfile);

# Why copy $mail to $mail2store? Because it is save. We have one for processing and one
# for the optional storage.

$mail2store = $mail;

# Lets detect whether it is a good old fashioned text mail or a a MIME type of scrap

$tom = $mail;
$tom =~ s/=.*$//iso;
$tom =~ s/^Mail::Audit:://iso;

# So now we extract some general fields

$head = $mail->head;
$sender = $head->get(From);
$subject = $head->get(Subject);
$subject =~ s/\\/\//g;

# Say goodbye to the linefeed

chomp($sender);
chomp($subject);

# Now we filter out the sender from the string delivered

$sender =~ m/\S+@\S+/;
$sender = $&;

# Now we have the sender but we must eliminate things like #,",< and >
# and convert all to lowercase letters (yes - MS is using uppercase ones)

$sender =~ s/^<//;
$sender =~ s/>$//;
$sender =~ s/^\'//;
$sender =~ s/\'$//;
$sender =~ s/^\"//;
$sender =~ s/\"$//;
$sender = lc "$sender";


# Begin separation of mail -----------------------------------------------------------------

# Here we have a flat mail. Thats easy.

if ($tom eq 'MailInternet')
   {
   chomp(my @body = @{$mail->body()});
   $content = join(" ",@body);
   }
   

# Now we reach we the real nightmare - MIME

if ($tom eq 'MimeEntity')
   {
   $bodyh   = $mail->bodyhandle;
   $type = $mail->mime_type;
   $num_parts = $mail->parts;
   $parts_counter="";
   $content_type = "";
   $tmpmail = "";

   $subject = decode_mimewords($subject,);
   $subject = to_ascii($subject);
   $subject =~ s/\\/\//g;


# So now we look if we have a bodyhandler. If not not we have more sh.t to do.

   if ($bodyh)
      {
      # We put the header in $header and replace all linefeeds with spaced
      # The we can filter out what we need
            
      $header = $mail->header_as_string();
      $header =~ s/\n+/ /g;
      
      # So first in which character set is it encoded?
      
      $charset = $header;    
      $charset =~ s/^.*Content-Type/Content-Type/g;
      $charset =~ s/^.*charset=//g;
      $charset =~ s/ .*$//g;

      # So second we determine the transfer type
      
      $content_transfer_encoding = get_transfer_encoding($header);


	  if ( $type eq "text/plain")
	     {
             $content = $mail->body_as_string();
             
             # If the content is base64 coded (like mails from MS MOM) decode it.

             if ( $content_transfer_encoding eq "base64")
	        {
                $content = decode_base64($content);
	        }

             # Now we remove unnecessary whitespace characters.
             $content = convert_unwanted($content);

	     }

	  if ( $type eq "text/html")
	     {

             # Now we convert it to a string
             $content = $mail->body_as_string();
             # If the content is base64 coded (like mails from MS MOM) decode it.

             if ( $content_transfer_encoding eq "base64")
	        {
                $content = decode_base64($content);
	        }

             
             # This little baby removes all HTML code
             $content =~ s/<(?:[^>'"]*|(['"]).*?\1)*>//gs;
             $content =~ s/&nbsp;+/ /g;
             
             # Now we remove unnecessary whitespace characters.
             $content = convert_unwanted($content);

             $content = to_ascii($content);

	     }
      }
   else
      {
      for ($parts_counter = 0; $parts_counter < $num_parts ; $parts_counter++)
          {

          $content_type = $mail->parts($parts_counter)->effective_type; 
          $parts_header = $mail->parts($parts_counter)->header_as_string();
          $parts_header =~ s/\n+/ /g;

          $content_transfer_encoding = get_transfer_encoding($parts_header);

	  if ( $content_type eq "text/plain")
	     {
	     $mail = MIME::Entity->build(Data => $mail->parts($parts_counter)->body_as_string);

             # Now we convert it to a string and remove unnecessary whitespace characters.

             $content = $mail->body_as_string();
             $content = convert_unwanted($content);

             last;
	     }

	  if ( $content_type eq "text/html")
	     {
	     $mail = MIME::Entity->build(Data => $mail->parts($parts_counter)->body_as_string);

             # Now we convert it to a string
             $content = $mail->body_as_string();
             
             # This little baby removes all HTML code
             $content =~ s/<(?:[^>'"]*|(['"]).*?\1)*>//gs;
             $content =~ s/&nbsp;+/ /g;

             # Now we remove unnecessary whitespace characters.
             $content = convert_unwanted($content);

             $content = to_ascii($content);
	     }
	  }
      }
   }

# End separation of mail -------------------------------------------------------------------


# Mail logging makes only sense when not in testmode because you need the logged mails for testing
# Logging mails that has been logged makes no sense

if (!$testmode)
   {
   log_mail();
   }
   

# Well now we should have 3 necessary parts: sender, subject and content. But we test it
# and in case we miss something we exit


# We won't accept mails without subject.

if(!$subject)
  {
  write_log("Error! No subject for $sender.");
  write_log("Sender: $sender");
  write_log("Content: $content");
  exit;
  }
   
# You NEED some content, don't you?

if(!$content)
  {
  write_log("Error! No content for $sender. Maybe empty or nested MIME");
  write_log("Sender: $sender");
  write_log("Subject: $subject");
  exit;
  }
   

# So now lets see if we can assign a host to the mailaddress

$sending_host = get_host($sender);
   

# So now lets see if we can assign a application to the mailaddress

$sending_application = get_application($sender);
   
# So now lets see if we have set multiline output or not

$multiline = get_multiline($sender);

if ($multiline == 1)
   {
   $multiline = "<br>";
   }
else
   {
   $multiline = $multiline_def;
   }

# Well - you definitely NEED a host entry

if(!$sending_host)
  {
  write_log("Error! No entry for $sender in $m2mon_cfg");
  write_log("Sender: $sender");
  write_log("Subject: $subject");
  write_log("Content: $content");
   
  exit;
  }


# And now we assemble the name of the file containing the appropriate filter

if($sending_application)
  {
  $filter=$filter_dir."/".$sending_host."_".$sending_application.".fltr";
  }
else
  {
  $filter=$filter_dir."/".$sending_host.".fltr";
  }
   
# Now we test if the filter exists and if the determined filter does not
# exists we use a default filter or (if no filter defined) exit

if(! open(MAIL_FILTER,"<$filter") )
  {
  if($filter_default)
    {
    write_log("Filter not found for $sender- using default filter");
    write_log("Sender: $sender");
    write_log("Subject: $subject");
    write_log("Content: $content");
    $filter = $filter_dir."/".$filter_default;
    }
    else
    {
    write_log("Filter not found for $sender. No default filter defined. Exiting.");
    write_log("Sender: $sender");
    write_log("Subject: $subject");
    write_log("Content: $content");
    exit;
    }
  }
else
  {
  # This is to be clean
  close(MAIL_FILTER);
  }


# Now we include the filter

require "$filter";

# And now, lo and behold, we call the filter

if($sending_application)
  {
  %filter_return=filter("$sender","$subject","$content","$sending_host","$multiline","$sending_application");
  }
else
  {
  %filter_return=filter("$sender","$subject","$content","$sending_host","$multiline");
  }


# You need service_state_id

if (!$filter_return{service_state_id})
   {
    write_log("Filter $filter delivered no valid or no service_state_id. Exiting.");
    write_log("Sender: $sender");
    write_log("Subject: $subject");
    write_log("Content: $content");
    exit;
   }

if ($testmode)
   {
   print "\n----------------------------------------------------------------------\n";
   print "Testoutput:\n";
   print "===========\n\n";

   print "Step 01: Sender\n";
   print "---------------\n";
   print "$sender\n\n";

   print "Step 02: Subject\n";
   print "----------------\n";
   print "$subject\n\n";

   print "Step 03: Content\n";
   print "----------------\n\n";
   print "$content\n\n";

   print "Step 04: Content (Filtered)\n";
   print "---------------------------\n";
   print "$filter_return{content}\n\n";

   print "Step 05: Host causing the error\n";
   print "-------------------------------\n";
   print "$filter_return{host}\n\n";

   print "Step 06: Servicestate ID\n";
   print "---------------------------\n";
   print "$filter_return{service_state_id}\n\n";

   print "Step 07: Service Desciption\n";
   print "----------------\n";
   print "$filter_return{service_desc}\n\n";
   print "----------------------------------------------------------------------\n";
   }
else
   {
   # Feed Nagios
   open(FIFO, ">> $nagios_cmd") or die "Couldn't open $nagios_cmd: $!\n";
#   print FIFO "[$epochtime] PROCESS_SERVICE_CHECK_RESULT;$filter_return{host};$filter_return{service_desc};$filter_return{service_state_id};$filter_return{content}\n";
   print FIFO "[$epochtime] ";
   print FIFO "PROCESS_SERVICE_CHECK_RESULT;";
   print FIFO "$filter_return{host};";
   print FIFO "$filter_return{service_desc};";
   print FIFO "$filter_return{service_state_id};";
   print FIFO "$filter_return{content}";
#   print FIFO "\n";
   close(FIFO);
   }

# And out
exit 0;

# Begin subroutines ------------------------------------------------------------------------

sub get_transfer_encoding
    {
    my $transfer_encoding = "";
    my @tmp_encoding;
    
    $transfer_encoding = $_[0];
    $transfer_encoding =~ s/^.*Content-Transfer-Encoding/Content-Transfer-Encoding/;

    @tmp_encoding = split / /, $transfer_encoding;

    $transfer_encoding = $tmp_encoding[1];

    return  $transfer_encoding;
    }

sub convert_unwanted
    {
    # This part of code removes carriage returns and linefeeds and converts 
    # tabs ans multiple space to on space. It also removes a trailing space and
    # (thanks to MS Windows) backslashes to slashes.
    
    my $tmp_msg = "";
    $tmp_msg = $_[0];
 
    $tmp_msg =~ s/\r+/ /g;
    $tmp_msg =~ s/\n+/ /g;
    $tmp_msg =~ s/\t+/ /g;
    $tmp_msg =~ s/ +/ /g;
    $tmp_msg =~ s/\\/\//g;
    $tmp_msg =~ s/^ //g;


    return  $tmp_msg;
    }


sub get_host()
    {

    # Let's open the configfile and process the content
    
    my $from_whom="$_[0]";
    my $host2get;

    $host2get = $config->val('Mail2Host', "$from_whom");
      
    return $host2get;
    }

sub get_application()
    {

    # Let's open the configfile and process the content
    
    my $from_whom="$_[0]";
    my $application2get;
        
    $application2get = $config->val('Mail2Appl', "$from_whom");
 
    return $application2get;
    }

sub get_multiline()
    {

    # Let's open the configfile and process the content
    
    my $from_whom="$_[0]";
    my $multi_or_not;
        
    $multi_or_not = $config->val('Multiline', "$from_whom");
 
    return $multi_or_not;
    }


sub ejection_seat()
    {
    my $msg2log="$_[0]";
    write_log("$msg2log");
    write_log("Sender: $sender");
    write_log("Subject: $subject");
    write_log("Content: $content");
    exit;
    }
    
sub write_log()
    {
    my $to_log="$_[0]";
    my $to_log_lenght=length "$to_log";
    
    if($content_logsize)
      {
      if($to_log_lenght > $content_logsize)
        {
        $to_log = substr($to_log,0,$content_logsize);
        }
      }
      
    # Timestamp for log
    (my $YEAR,my $MONTH,my $DAY, my $HOUR,my $MINUTE,my $SEC) = Time_to_Date(time());
    my $logstamp = sprintf("%.3s %02d %02d:%02d:%02d",Month_to_Text($MONTH),$DAY,$HOUR,$MINUTE,$SEC);

    if ($testmode)
       {
       print "$logstamp $to_log\n";
       }
    else
       {
       open(LOGFILE,">>$logfile") || die "Can not open logfile";
    
       print LOGFILE "$logstamp $to_log\n";
    
       close(LOGFILE);
       }
    }
    
sub log_mail()
    {
    
    # So now we get the entries from configfile
    
    $logmail{store_mail}  = $config->val('LogMail', 'store_mail');
    $logmail{mail_store}  = $config->val('LogMail', 'mail_store');
    
    # And if enabled we store the mails in the mailstore directory as seperate files
    
    if( $logmail{store_mail} == 1 )
      {
      mkdir "$logmail{mail_store}";
      $mail2store->accept($logmail{mail_store}, { one_for_all => 1, noexit => 1});
      }
    }


sub print_usage
    {
    print "\nUsage: $progname -t|--test -l|--logfile <logfile>\n\n";
    print "or\n";
    print "\nUsage: $progname -h for help.\n\n";
    }


sub print_help
    {
    print "Copyright (c) 2009 Martin Fuerstenau\n";
    print_usage();
    print "       -t, --test                                    Testmode - Will print the output of the filtered\n";
    print "                                                     mail to stdout. For testing new filters.\n";
    print "\n";
    print "       -h, --help                                    Short help message (mandatory)\n";
    print "\n";
    print "       -l, --logfile                                 Logfile location (optional)\n";
    print "\n";
    }
